from pathlib import Path

from django.conf import settings
from django.template.defaulttags import (
    ForNode,
    IfNode,
)
from django.template.loader_tags import IncludeNode
from django.templatetags.i18n import (
    BlockTranslateNode,
    TranslateNode,
)

from iommi._web_compat import format_html
from iommi.debug import src_debug_url_builder
from iommi.thread_locals import get_current_request


def setup_code_finder():
    if not settings.DEBUG:
        return

    for node_class in [IncludeNode, ForNode, IfNode, TranslateNode, BlockTranslateNode]:
        def debug_url_render(self, context, orig_render=node_class.render):
            orig_result = orig_render(self, context)
            request = get_current_request()
            if not request or '_iommi_code_finder' not in request.GET:
                return orig_result

            # Get the file path and line number from the node's origin and token
            link_html = ''
            if hasattr(self, 'origin') and self.origin:
                filename = self.origin.name
                # Token has a lineno attribute that gives us the line number
                line_number = self.token.lineno if hasattr(self, 'token') and hasattr(self.token, 'lineno') else 1

                # Create a PyCharm link
                link_html = format_html(
                    '<!-- ## iommi-code-finder-URL ## {} ## {} -->',
                    f'{Path(filename).name}:{line_number}',
                    src_debug_url_builder(filename, line_number)
                )

            result = link_html + orig_result
            return result

        node_class.render = debug_url_render



from iommi import (
    Asset,
    Style,
)
from iommi.style_bootstrap5 import bootstrap5

bootstrap_docs = Style(
    bootstrap5,
    root__assets__doc_style=Asset.css(attrs__href='/_static/iframe_custom.css'),
    root__assets__doc_js=Asset.js(attrs__src='/_static/iframe_custom.js'),
    root__assets__iommi_js=Asset.js(attrs=dict(src='/_static/js/iommi.js')),
    root__assets__iommi_code_finder_js=Asset.js(
        attrs=dict(src='/_static/js/iommi_code_finder.js'),
        include=lambda request, **_: '_iommi_code_finder' in request.GET if request else False,
    ),
    root__assets__iommi_css=Asset.css(attrs__href='/_static/css/iommi.css'),
    root__assets__iommi_scroll_js=Asset.js(attrs__src='/_static/js/iommi-scroll.js'),
    root__assets__iommi_menu_css=Asset.css(attrs__href='/_static/css/iommi_main_menu.css'),

    internal=True,
    Container=dict(
        attrs__class={
            'container': False,
            'mt-5': False,
        },
        attrs__style__padding='1rem',
    ),
    Admin__parts__table__h_tag__attrs__style={'margin-top': '3rem'},

    MainMenu=dict(
        assets=dict(
            iommi_main_menu_css=Asset.css(attrs__href='/_static/css/iommi_main_menu.css')
        ),
    ),
)



from functools import cached_property
from pathlib import PurePosixPath
from urllib.parse import urlparse

from django.conf import settings
from django.conf.urls import handler403
from django.core.exceptions import PermissionDenied
from django.urls import (
    get_resolver,
    include,
    path as orig_path,
)
from django.utils.translation import gettext_lazy

from iommi import (
    MISSING,
)
from iommi._web_compat import (
    format_html,
    render_template,
)
from iommi.attrs import (
    evaluate_attrs,
    render_attrs,
)
from iommi.base import capitalize
from iommi.declarative.dispatch import dispatch
from iommi.declarative.namespace import (
    EMPTY,
    Namespace,
)
from iommi.evaluate import (
    evaluate_as_needed,
    evaluate_strict,
)
from iommi.path import decode_path_components
from iommi.struct import Struct
from iommi.style import resolve_style

EXTERNAL = object()


class NoTranslateString(str):
    pass


def path(path, view_or_list, kwargs=None):
    if isinstance(view_or_list, list):
        assert kwargs is None
        return orig_path(path, include(view_or_list))
    elif isinstance(view_or_list, type):
        return orig_path(path, view_or_list().as_view(), kwargs=kwargs)
    else:
        try:
            return orig_path(path, view_or_list.as_view(), kwargs=kwargs)
        except AttributeError:
            return orig_path(path, view_or_list, kwargs=kwargs)


class MainMenu:
    @dispatch(
        items=EMPTY,
        attrs__style=EMPTY,
        attrs__class=EMPTY,
        extra=EMPTY,
        extra_evaluated=EMPTY,
    )
    def __init__(self, *, items=None, attrs, template=MISSING, paths=None, extra, extra_evaluated):
        self.items = items
        self.attrs = attrs
        self.template = template
        self.paths = paths or []
        self.extra = extra
        self.extra_evaluated = extra_evaluated
        for name, c in self.items.items():
            c.parent = self
            c._set_name(name)

    def urlpatterns(self):
        if 'iommi.experimental.main_menu.main_menu_middleware' not in settings.MIDDLEWARE:
            # This is an explicit raise instead of an `assert` because I'm paranoid of people running Python in optimized mode with skipped asserts.
            raise AssertionError('To use the iommi main menu, you must add `iommi.experimental.main_menu.main_menu_middleware` to the MIDDLEWARE list.')

        return [
            path(x.path, x.urlpatterns())
            for x in self.items.values()
            if x.view is not EXTERNAL
        ] + self.paths

    def bind(self, request):
        style = resolve_style(None)
        conf = Namespace(*style.resolve(self))

        template = conf.pop('template', 'iommi/main_menu/menu.html')
        if self.template is MISSING:
            template = template

        assets = {
            k: v.bind(request=request)
            for k, v in conf.pop('assets', {}).items()
        }

        attrs = Namespace(self.attrs, conf.pop('attrs', {}))

        assert not conf, f'Unknown configuration {conf} for `MainMenu`'

        return BoundMainMenu(
            self,
            request,
            attrs=attrs,
            template=template,
            assets=assets,
        )

    def __repr__(self):
        return '<MainMenu>'


class BoundMainMenu:
    def __init__(self, main_menu, request, attrs, template, assets):
        self.main_menu = main_menu
        self.request = request
        self.attrs = attrs
        self.template = template
        self.paths = main_menu.paths
        self.assets = assets
        self._own_evaluate_parameters = self.own_evaluate_parameters()

        self.attrs = evaluate_attrs(self, **self._own_evaluate_parameters)

        self.raw_items = {
            k: v.bind(request=request, root=self)
            for k, v in self.main_menu.items.items()
        }
        self.items = {k: v for k, v in self.raw_items.items() if v.include}

        # Set active_item
        self.url = None
        best_match = None
        path_parts = PurePosixPath(urlparse(self.request.get_full_path()).path).parts

        def url_likeness(x):
            if x is None:
                return 0
            a = PurePosixPath(x.url).parts
            b = path_parts
            # if a starts with b
            if a == b[:len(a)]:
                return len(a)
            return 0

        def foo(x):
            nonlocal best_match

            if not x.params_are_satisfied():
                return

            if x.url and '://' not in x.url and url_likeness(best_match) < url_likeness(x):
                best_match = x

            for sub_item in x.raw_items.values():
                foo(sub_item)

        foo(self)
        self.active_item = best_match

    def check_access(self):
        for item in self.raw_items.values():
            if not item.is_active():
                continue
            item.check_access()

    def params_are_satisfied(self):
        return True

    def __str__(self):
        return self.__html__()

    def __html__(self):
        return render_template(self.request, self.template, self._own_evaluate_parameters)

    def render_items(self):
        return format_html('{}' * len(self.items), *[x.__html__() for x in self.items.values()])

    @property
    def extra(self):
        return self.main_menu.extra

    @cached_property
    def extra_evaluated(self):
        return Struct(evaluate_as_needed(self.main_menu.extra_evaluated or {}, self._own_evaluate_parameters))

    def __repr__(self):
        return '<BoundMainMenu>'

    def own_evaluate_parameters(self):
        request = self.request
        return {
            'request': request,
            'user': getattr(request, 'user', None) if request else None,
            'main_menu': self,
            **getattr(request, 'iommi_view_params', {}),
        }


class M:
    @dispatch(
        items=EMPTY,
        attrs__style=EMPTY,
        attrs__class=EMPTY,
        extra=EMPTY,
        extra_evaluated=EMPTY,
    )
    def __init__(self, *, icon=None, view=None, view_kwargs=None, path=MISSING, url=None, open=None, params=None, display_name=None, items=None, include=None, attrs, template=MISSING, paths=None, extra, extra_evaluated, render=True):

        assert view is not None or view is EXTERNAL, f'Items should either have a view function specified, or supply EXTERNAL to mark the item as having an external URL and thus not needing access control. Got {view!r}'

        self.name = None
        self.icon = icon
        self.render = render
        self.url = url
        assert url is None or isinstance(url, str) or callable(url)
        self.path = path
        assert path is MISSING or path is None or isinstance(path, str), f'`path` must be `MISSING`, `None` or a `str`. Got {type(self.path)} ({self.path}).'
        self.paths = paths or []
        assert isinstance(self.paths, list), f'`paths` must be of type list. Got type {type(self.paths)} ({self.paths}).'
        self.view = view
        self.view_kwargs = view_kwargs
        self.include = include
        self.display_name = display_name
        self._raw_display_name = None  # We need programmatic access to this from okrand for i18n
        self.dynamic_items = None
        self.items = {}

        if isinstance(items, Namespace) and 'call_target' in items:
            self.dynamic_items = items.pop('call_target')
            assert not items, "You can't mix dynamic items and static items"
        else:
            self.items = {
                k: M(**v) if isinstance(v, dict) else v
                for k, v in items.items()
            }

        self.parent = None
        self.params = params
        self.open = open
        self.attrs = attrs
        self.template = template
        self.extra = extra
        self.extra_evaluated = extra_evaluated
        if params is not None:
            assert isinstance(params, set), '`params` must be a `set`'
            assert self.path, 'if `params` is set, `path` must also be set'
        for name, c in self.items.items():
            c.parent = self
            c._set_name(name)

        if isinstance(self.url, str) and self.path is MISSING:
            assert not self.items and not self.paths, "External links can't have subitems"

        if self.view is EXTERNAL:
            assert not self.items, 'External views are not allowed to have subitems. Use a redirect view instead.'

    def __repr__(self):
        return f"<M {self.name}>"

    def _set_name(self, name):
        self.name = name
        if not callable(self.display_name) and not isinstance(self.display_name, NoTranslateString):
            self._raw_display_name = (self.display_name or name).replace('_', ' ')
            self.display_name = capitalize(gettext_lazy(self._raw_display_name))
        if self.path is MISSING:
            self.path = self.name + '/'

    def urlpatterns(self):
        if isinstance(self.url, str) and self.path is MISSING:
            return []

        if self.view is EXTERNAL:
            return []

        return [
            path('', self.view, kwargs=self.view_kwargs)
        ] + [
            path(x.path, x.urlpatterns())
            for x in self.items.values()
            if x.view is not EXTERNAL
        ] + self.paths

    def bind(self, request, root):
        return BoundM(
            self,
            request=request,
            root=root,
            parent=None,
        )


class BoundM:
    def __init__(self, m, *, request, parent, root):
        self.m = m
        self.request = request
        self.parent = parent
        self.root = root
        self.include = self._include()

        style = resolve_style(None)
        conf = Namespace(*style.resolve(self))

        template = conf.pop('template', 'iommi/main_menu/menu_item.html')
        if self.m.template is not MISSING:
            template = self.m.template

        self.template = template
        attrs = Namespace(self.m.attrs, conf.pop('attrs', {}))
        self.attrs = attrs

        assert not conf, f'Unsupported configuration {conf}'

        self._own_evaluate_parameters = self.own_evaluate_parameters()

        self.render_item = evaluate_strict(self.m.render, **self._own_evaluate_parameters)

        if self.include:
            self.display_name = self._display_name()
            items = self.m.items

            if self.m.dynamic_items:
                assert not items
                items = evaluate_strict(self.m.dynamic_items, **self._own_evaluate_parameters)

            self.raw_items = {
                k: BoundM(
                    v,
                    request=request,
                    root=root,
                    parent=self,
                )
                for k, v in items.items()
            }
            self.items = {k: v for k, v in self.raw_items.items() if v.include}
        else:
            self.raw_items = {}
            self.items = {}

        self.attrs = evaluate_attrs(self, **self._own_evaluate_parameters)

        self.has_rendered_items = len([x for x in self.items.values() if x.render_item]) > 0

    def __str__(self):
        return self.__html__()

    def __html__(self):
        if not self.render_item:
            return ''
        return self.render()

    def render(self):
        return render_template(self.request, self.template, context=self.own_evaluate_parameters())

    def __repr__(self):
        return f'<BoundM {self.m.name}>'

    def is_active(self):
        return self.root.active_item is self or any([item.is_active() for item in self.raw_items.values()])

    def own_evaluate_parameters(self):
        request = self.request
        return {
            'request': request,
            'user': request.user if request else None,
            'item': self,
            'main_menu': self.root,
            **getattr(request, 'iommi_view_params', {}),
        }

    @cached_property
    def rendered_attrs(self):
        if self.is_active():
            self.attrs['class']['is_active'] = True
        return render_attrs(self.attrs)

    @cached_property
    def open(self):
        if self.m.open is not None:
            return evaluate_strict(self.m.open, **self.own_evaluate_parameters())

        return self.is_active()

    def params_are_satisfied(self):
        assert hasattr(self.request, 'resolver_match')
        if self.m.params:
            return len(self.m.params & set(getattr(self.request, 'iommi_view_params', {}).keys())) == len(self.m.params)
        else:
            return True

    def _include(self):
        if self.m.include is not None:
            return evaluate_strict(self.m.include, **self.own_evaluate_parameters())
        return self.params_are_satisfied()

    @cached_property
    def icon(self):
        return evaluate_strict(self.m.icon, **self.own_evaluate_parameters())

    @cached_property
    def link(self):
        url = self.url
        icon = ''
        external = ''
        target = ''

        style = resolve_style(None)

        if self.icon is not None:
            icon = style.icon_formatter(self.icon)

        if url.startswith('https://') or url.startswith('http://'):
            external = format_html(' {} ', style.icon_formatter('external'))
            target = ' target="blank"'
        return format_html('<a href="{}" title="{}" {}>{}<span>{}{}</span></a>', url, self.display_name, target, icon, self.display_name, external)

    @property
    def extra(self):
        return self.m.extra

    @cached_property
    def extra_evaluated(self):
        return Struct(evaluate_as_needed(self.m.extra_evaluated or {}, self._own_evaluate_parameters))

    @cached_property
    def url(self):
        if self.m.url is None:
            if self.parent is not None and self.parent.url is not None:
                assert self.parent.url.endswith('/')
                return self.parent.url + self.m.name + '/'
            else:
                return '/' + self.m.name + '/'

        return evaluate_strict(self.m.url, **self.own_evaluate_parameters())

    def _display_name(self):
        return evaluate_strict(self.m.display_name, **self.own_evaluate_parameters())

    def check_access(self):
        if self.is_active() and not self.include:
            raise PermissionDenied()
        for item in self.raw_items.values():
            if not item.is_active():
                continue
            item.check_access()


def main_menu_middleware(get_response):
    def main_menu_middleware_inner(request):
        if request.resolver_match is None:
            request.resolver_match = get_resolver().resolve(request.path_info)

        if request.resolver_match.app_name != 'ninja':
            if not hasattr(request, 'iommi_view_params'):
                try:
                    if not hasattr(request, 'iommi_view_params'):
                        request.iommi_view_params = Struct()
                    request.iommi_view_params.update(**decode_path_components(request, **request.resolver_match.kwargs))
                except PermissionDenied as exception:
                    return handler403(request, exception=exception)

        menu_fully_qualified_name = getattr(settings, 'IOMMI_MAIN_MENU', None)
        assert menu_fully_qualified_name is not None, "To use the main_menu_middleware, you must define settings.IOMMI_MAIN_MENU as a string to the full path to the main menu declaration"
        assert '.' in menu_fully_qualified_name, "IOMMI_MAIN_MENU must be in `your_module.symbol` format"
        menu_module_name, _, menu_symbol_name = menu_fully_qualified_name.rpartition('.')

        module = __import__(menu_module_name, fromlist=['_silly_importlib'])
        menu = getattr(module, menu_symbol_name)
        request.iommi_main_menu = menu.bind(request=request)

        try:
            request.iommi_main_menu.check_access()
        except PermissionDenied as exception:
            return handler403(request, exception=exception)

        return get_response(request)

    return main_menu_middleware_inner



import inspect
import re
from collections import defaultdict
from glob import glob
from os.path import join
from pathlib import Path
from textwrap import (
    dedent,
    indent,
)
from typing import get_type_hints

from iommi import (
    MISSING,
    Part,
)
from iommi.base import items
from iommi.declarative import get_declared
from iommi.declarative.namespace import (
    flatten,
    Namespace,
)
from iommi.refinable import (
    EvaluatedRefinable,
    is_refinable_function,
    SpecialEvaluatedRefinable,
)
from iommi.shortcut import (
    get_shortcuts_by_name,
    is_shortcut,
)
from iommi.struct import merged
from iommi.struct import Struct  # noqa: E402
from tests.helpers import create_iframe  # noqa: E402


def uses_from_cookbooks():
    import docutils.parsers.rst
    import docutils.utils
    import docutils.frontend
    from docutils.nodes import comment, title, target, section
    from iommi.declarative.util import strip_prefix

    parser = docutils.parsers.rst.Parser()
    components = (docutils.parsers.rst.Parser,)
    settings = docutils.frontend.OptionParser(components=components).get_default_values()

    backrefs = defaultdict(set)

    for filename in glob(str(Path(__file__).parent.parent / 'docs' / 'cookbook_*.rst')):
        with open(filename) as f:
            content = f.read()

        document = docutils.utils.new_document(filename, settings=settings)
        parser.parse(content, document)

        targets = [x for x in document.children if isinstance(x, target)]
        nodes = [x for x in document.children if not isinstance(x, target)]

        target_node = None if not targets else targets[0]

        assert len(nodes) == 1, 'Expected a document with only ONE level 1 header'
        (d,) = nodes

        for node in d.children:
            if isinstance(node, target):
                target_node = node
                continue
            if not node.attributes['ids']:
                continue
            if not node.children:
                continue
            if not isinstance(node, section):
                continue
            title_node = node.children[0]
            assert isinstance(title_node, title)

            for c in node.children:
                x = c.astext()

                if isinstance(c, target):
                    target_node = c

                elif isinstance(c, comment) and x.startswith('uses '):
                    if x.endswith(':'):
                        print(f'WARNING: bad `uses`: {x}')
                    if not target_node:
                        print(f'WARNING: no target for {title_node.astext()}. It must be ABOVE the title.')
                        continue

                    backrefs[strip_prefix(x, prefix='uses ')].add((target_node.attributes['ids'][0], title_node.astext()))

    assert backrefs
    return backrefs


def get_default_classes():
    import iommi

    return [
        iommi.Table,
        iommi.Column,
        iommi.EditTable,
        iommi.EditColumn,
        iommi.Query,
        iommi.Filter,
        iommi.Form,
        iommi.fragment.Fragment,
        iommi.Field,
        iommi.Action,
        iommi.Page,
        iommi.Menu,
        iommi.MenuItem,
        iommi.Style,
        iommi.fragment.Header,
        iommi.Asset,
        iommi.middleware,
        # Private-ish APIs
        iommi.endpoint.Endpoint,
        iommi.part.Part,
        iommi.traversable.Traversable,
        iommi.member.Members,
        iommi.table.Cell,
        iommi.table.Cells,
        iommi.table.ColumnHeader,
        iommi.table.HeaderConfig,
        iommi.attrs.Attrs,
        iommi.table.ColumnHeader,
        iommi.fragment.Container,
        iommi.table.TableAutoConfig,
        iommi.form.FormAutoConfig,
        iommi.query.QueryAutoConfig,
    ]


def generate_api_docs_tests(directory, classes=None, verbose=False):  # pragma: no cover - this is tested by rtd anyway
    """
    Generate test files for declarative APIs

    :param directory: directory to write the .py files into
    :param classes: list of classes to generate tests for
    :param verbose: print verbose warnings for missing docs
    """
    if classes is None:
        classes = get_default_classes()

    doc_by_filename = _generate_tests_from_class_docs(classes=classes, verbose=verbose)  # pragma: no mutate
    for source_filename, filename, doc_generator in doc_by_filename:  # pragma: no mutate
        doc = doc_generator()
        with open(directory / filename, 'w') as f2:  # pragma: no mutate
            f2.write(doc)  # pragma: no mutate


def get_docs_callable_description(c):
    if getattr(c, '__name__', None) == '<lambda>':
        import inspect

        return inspect.getsource(c).strip()
    return c.__module__ + '.' + c.__name__


def get_methods_by_type_by_name(class_):
    function_type = type(get_methods_by_type_by_name)
    ignore_list = ['get_declared', 'set_declared', 'get_meta']

    r = {
        k: v
        for k, v in class_.__dict__.items()
        if not k.startswith('_')
        and k not in ignore_list
        and not is_shortcut(getattr(class_, k))
        and not is_refinable_function(getattr(class_, k))
    }

    return {
        'Methods': {k: v for k, v in r.items() if isinstance(v, function_type)},
        'Static methods': {k: v for k, v in r.items() if isinstance(v, staticmethod)},
        'Class methods': {k: v for k, v in r.items() if isinstance(v, classmethod)},
    }


def docstring_param_dict(obj):
    doc = obj.__doc__
    if doc is None:
        return dict(text=None, params={})
    doc = dedent(doc)
    return dict(
        text=doc[: doc.find(':param')].strip() if ':param' in doc else doc.strip(),
        params=dict(re.findall(r":param (?P<name>\w+): (?P<text>.*)", doc)),
    )


def indent_levels(levels, s):
    return (' ' * levels * 4) + s


def get_namespace(c):
    return Namespace(
        {
            k: getattr(c.__init__, '__iommi_with_defaults_kwargs', {}).get(k)
            for k, v in items(get_declared(c, 'refinable'))
        }
    )


def _generate_tests_from_class_docs(classes, verbose=False):
    uses_by_field = uses_from_cookbooks()

    for c in classes:
        from io import StringIO

        f = StringIO()
        yield _generate_tests_from_class_doc(f, c, classes, uses_by_field, verbose=verbose)


def _generate_tests_from_class_doc(f, c, classes, uses_by_field, verbose=False):
    return (
        inspect.getfile(c),
        'test_doc__api_%s.py' % c.__name__,
        lambda: _generate_tests_from_class_doc_inner(f, c, classes, uses_by_field, verbose=verbose),
    )


concepts = {
    'after': 'after',
    'assets': 'assets',
    'attr': 'attr',
    'attrs': 'attributes',
    'auto': 'auto',
    'endpoints': 'endpoints',
    'extra': 'extra',
    'extra_evaluated': 'extra',
    'extra_params': 'extra_params',
    'include': 'include',
    'iommi_style': 'iommi_style',
    'name': 'name',
    'display_name': 'name',
    'tag': 'tag',
    'template': 'template',
    'title': 'title',
    'h_tag': 'title',
}


def _generate_tests_from_class_doc_inner(f, c, classes, uses_by_field, verbose):
    def w(levels, s):
        f.write(indent_levels(levels, s))
        f.write('\n')

    def section(level, title, indent=0):
        underline = {0: '=', 1: '-', 2: '^', 3: '+'}[level] * len(title)
        w(indent, title)
        w(indent, underline)
        w(indent, '')

    w(
        0,
        '''
# NOTE: this file is automatically generated

from iommi import *
from iommi.admin import Admin
from iommi.struct import Struct
from django.urls import (
    include,
    path,
)
import pytest
from django.db import models
from tests.helpers import req, user_req, staff_req, show_output
from docs.models import *

pytestmark = pytest.mark.django_db

@pytest.fixture(autouse=True)
def auto_use(big_discography):
    pass

request = req('get')


# language=rst
"""
    ''',
    )

    section(0, c.__name__, indent=0)

    class_text = ''
    constructor_text = ''
    params = {}
    for x in reversed(c.__mro__):
        foo = docstring_param_dict(x.__init__)
        constructor_text = foo.pop('text')
        params = merged(foo.pop('params'), params)

        foo = docstring_param_dict(x)
        class_text = foo.pop('text')
        params = merged(foo.pop('params'), params)
        assert not foo

    if c.__base__ in classes:
        w(0, f'Base class: :doc:`{c.__base__.__name__}`')
    else:
        w(0, f'Base class: `{c.__base__.__name__}`')

    w(0, '')

    w(0, '"""')
    w(0, 'def test_base():')
    w(1, '# language=rst')
    w(1, '"""')

    _print_rst_or_python(class_text, w)

    w(1, '"""')
    w(0, '')
    w(1, '# language=rst')
    w(1, '"""')

    if constructor_text:
        if constructor_text:
            w(0, '')

        f.write(constructor_text)
        w(0, '')

    w(0, '')

    defaults = Namespace()
    for refinable, value in sorted(get_namespace(c).items()):
        if value not in (None, MISSING):
            defaults[refinable] = value

    def default_description(v):
        if callable(v) and not isinstance(v, Namespace):
            v = get_docs_callable_description(v)

            if 'lambda' in v:
                v = v[v.find('lambda'):]
                v = v.strip().strip(',').replace('\n', ' ').replace('  ', ' ')
        if isinstance(v, Part):
            v = v.bind()
        if v == '':
            v = '""'
        return v

    refinable_members = sorted(dict.items(get_namespace(c)))
    if refinable_members:
        section(1, 'Refinable members')
        type_hints = get_type_hints(c)
        for refinable, value in refinable_members:
            evaluated_marker = ''
            if isinstance(getattr(c, refinable, None), (EvaluatedRefinable, SpecialEvaluatedRefinable)):
                evaluated_marker = (
                    '  \N{NO-BREAK SPACE}\N{NO-BREAK SPACE}\N{NO-BREAK SPACE}  (:ref:`evaluated <evaluate>`)'
                )

            w(0, '')
            section(2, '`' + refinable + '`' + evaluated_marker)

            docstring = getattr(getattr(c, refinable, None), '__doc__')
            if docstring:
                _print_rst_or_python(docstring, w, indent=0)
                w(0, '')
            elif params.get(refinable):
                w(0, params[refinable])
                w(0, '')
            type_hint = type_hints.get(refinable)
            if type_hint:
                name = str(type_hint)
                if name.startswith('typing.'):
                    name = name.replace('typing.', '')
                else:
                    name = type_hint.__name__

                if type_hint in classes:
                    w(0, f'Type: :doc:`{name}`')
                else:
                    w(0, f'Type: `{name}`')
                w(0, '')

            if refinable in defaults:
                w(0, f'Default: `{default_description(defaults.pop(refinable))}`')

            if refinable in concepts:
                w(1, f'See :ref:`{concepts[refinable]} <{concepts[refinable]}>`')
                w(0, '')

            cookbook_usages = uses_by_field.get(f'{c.__name__}.{refinable}', [])
            if cookbook_usages:
                w(0, '')
                w(0, f'Cookbook:')
                for id_, title in cookbook_usages:
                    w(1, f':ref:`{id_}`')
                    w(0, '')
            else:
                if refinable not in concepts and verbose:
                    print(f'WARNING: {c.__name__}.{refinable} has no cookbook examples')

        w(0, '')

    assert not defaults

    shortcuts = get_shortcuts_by_name(c)
    if shortcuts:
        section(1, 'Shortcuts')

        for name, shortcut in sorted(shortcuts.items()):
            section(2, f'`{c.__name__}.{name}`')

            if shortcut.__doc__:
                foo = docstring_param_dict(shortcut)
                _print_rst_or_python(foo['text'], w)
                w(0, '')

                if foo['params']:
                    w(0, '')
                    section(3, 'Parameters')
                    for k, v in foo['params'].items():
                        w(0, f'* `{k}`')
                        w(1, f'* `{v}`')

                w(0, '')

            try:
                instance = shortcut()
            except Exception:
                pass  # todo For now...
            else:
                parents = instance.iommi_shortcut_stack[1:]
                if parents:
                    parent = parents[0]

                    w(0, f'Parent: {c.__name__}.{parent}_')
                    w(0, '')

            defaults = shortcut if isinstance(shortcut, dict) else getattr(shortcut, '__iommi_with_defaults_kwargs', {})
            if defaults:
                defaults = Namespace(defaults)
                section(3, 'Defaults')
                for k, v in items(flatten(defaults)):
                    v = default_description(v)
                    w(0, f'* `{k}`')
                    w(1, f'* `{v}`')
                w(0, '')

            cookbook_usages = uses_by_field.get(f'{c.__name__}.{name}', [])
            if cookbook_usages:
                w(0, '')
                w(0, f'Cookbook:')
                for id_, title in cookbook_usages:
                    w(1, f':ref:`{id_}`')
                    w(0, '')

    for k, v in get_methods_by_type_by_name(c).items():
        methods = v
        if methods:
            section(1, k)

            for name, method in sorted(methods.items()):
                section(2, f'`{name}`')

                doc = getattr(c, name).__doc__
                if doc:
                    w(0, '')
                    _print_rst_or_python(doc, w)
                    w(0, '')

    w(1, '''"""''')
    return f.getvalue()


def _print_rst_or_python(doc, w, indent=0):
    if not doc:
        return
    in_code_block = False
    code_block_indent = None
    for line in dedent(doc).split('\n'):
        assert '\n' not in line
        if line.strip().startswith('.. code-block:: python'):
            w(1, '"""')

            in_code_block = True
        elif in_code_block:
            if code_block_indent is None:
                if line.strip():
                    code_block_indent = len(line) - len(line.lstrip(' '))
                w(0, line)
            else:
                if line.startswith(' ' * code_block_indent) or not line.strip():
                    w(0, line)
                else:
                    in_code_block = False
                    w(1, '# language=rst')
                    w(1, '"""')
                    w(1 + indent, line)
        else:
            w(0 + indent, line)
    if in_code_block:
        w(1, '# language=rst')
        w(1, '"""')
    w(0, '')


def write_rst_from_pytest():
    for source in (Path(__file__).parent.parent / 'docs/').glob('test_*.py'):
        target = source.parent / f'{source.stem.replace("test_doc__api_", "").replace("test_doc_", "")}.rst'

        with open(source) as source_f:
            with open(target, 'w') as target_f:
                rst_from_pytest(source_f, target_f, target)


def rst_from_pytest(source_f, target_f, target):
    blocks = []
    stack = []
    state_ = None

    def push(new_state, **kwargs):
        nonlocal state_
        nonlocal i
        assert state_ != 'only test', ('exited @test without @end', source_f.name, i)
        stack.append(new_state)
        blocks.append(Struct(state=new_state, lines=[], metadata=kwargs))
        state_ = new_state

    def pop():
        nonlocal state_
        stack.pop()
        state_ = stack[-1]
        blocks.append(Struct(state=state_, lines=[], metadata={}))

    def add_line(line):
        blocks[-1].lines.append(line)

    push('import')

    func_name = None
    func_count = 0

    for i, line in enumerate(source_f.readlines(), start=1):
        stripped_line = line.strip()
        if state_ in ('import', 'py') and line.startswith('def test_'):  # not stripped_line!
            func_name = line[len('def ') :].partition('(')[0]
            push('py', func_name=func_name, func_count=0)
            func_count = 0
        elif stripped_line.startswith("# language=rst"):
            push('starting rst')
        elif stripped_line in ('"""', "'''"):
            if state_ == 'starting rst':
                # add_line('')
                pop()
                push('rst')
            elif state_ == 'rst':
                pop()
        elif stripped_line.startswith('# @test'):
            push('only test')
        elif stripped_line.startswith('# @end'):
            pop()
        elif state_ == 'py' and line.startswith(
            '#'
        ):  # not stripped_line! skip comments on the global level between functions
            continue
        elif state_ == 'py' and line.startswith(
            '@'
        ):  # not stripped_line! skip decorators on the global level between functions
            continue
        else:
            if state_ == 'only test':
                if stripped_line.startswith('show_output(') or stripped_line.startswith('show_output_collapsed('):
                    name = join(target.stem, func_name)
                    if func_count:
                        name += str(func_count)
                    func_count += 1

                    blocks.append(
                        Struct(
                            state='raw',
                            lines=[create_iframe(name, collapsed=stripped_line.startswith('show_output_collapsed'))],
                            metadata={},
                        )
                    )
            else:
                add_line(line)

    for b in blocks:
        b.text = dedent(''.join(b.lines)).strip()
        del b['lines']

    blocks = [x for x in blocks if x.text]

    for b in blocks:
        if b.state == 'rst':
            target_f.write(b.text)
        elif b.state == 'py':
            target_f.write('.. code-block:: python\n\n')
            target_f.write(indent(b.text, '    '))
        elif b.state == 'raw':
            target_f.write('.. raw:: html\n\n')
            target_f.write(indent(b.text, '    '))

        target_f.write('\n\n')



# Based on https://www.djangosnippets.org/snippets/186/

import cProfile
import os
import pstats
import subprocess
import sys
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile

from django.http.response import (
    HttpResponseBase,
    StreamingHttpResponse,
)
from django.utils.html import escape

from iommi._web_compat import (
    HttpResponse,
    settings,
)
from iommi.debug import src_debug_url_builder


MEDIA_PREFIXES = ['/static/']

_dot_search_paths = [
    '/usr/bin/dot',
    '/usr/local/bin/dot',
]


def get_dot_path():
    for p in _dot_search_paths:
        if os.path.exists(p):
            return p
    return None


def should_profile(request):
    disabled = getattr(request, 'profiler_disabled', True)
    is_staff = hasattr(request, 'user') and request.user.is_staff

    return ('_iommi_prof' in request.GET or '_iommi_prof' in request.POST) and (
        (not disabled and is_staff) or settings.DEBUG
    )


def strip_extra_path(s, token):
    if token not in s:
        return s
    pre, _, post = s.rpartition(' ')
    post = post[post.rindex(token) + len(token) :]
    return f'{pre} {post}'


class HTMLStats(pstats.Stats):
    def print_title(self):
        print(
            # language=HTML
            '''
                <thead>
                    <tr>
                        <th class="numeric"><a href="?_iommi_prof=ncalls">ncalls</a></th>
                        <th class="numeric"><a href="?_iommi_prof=tottime">tottime</a></th>
                        <th class="numeric">percall</th>
                        <th class="numeric"><a href="?_iommi_prof=cumtime">cumtime</a></th>
                        <th class="numeric">percall</th>
                        <th>function</th>
                        <th></th>
                        <th>filename</th>
                        <th>lineno</th>
                    </tr>
                </thead>
            ''',
            file=self.stream,
        )

    def print_stats(self, *amount):
        for filename in self.files:
            print(filename, file=self.stream)
        if self.files:
            print(file=self.stream)
        indent = ' ' * 8
        for func in self.top_level:
            print(indent, func[2], file=self.stream)

        print(indent, self.total_calls, "function calls", end=' ', file=self.stream)
        if self.total_calls != self.prim_calls:
            print("(%d primitive calls)" % self.prim_calls, end=' ', file=self.stream)
        print("in %.3f seconds" % self.total_tt, file=self.stream)
        print(file=self.stream)

        # this call prints...
        width, list = self.get_print_list(amount)

        print('<table>', file=self.stream)
        if list:
            self.print_title()
            limit = 280
            for func in list[:limit]:
                self.print_line(func)
            print(file=self.stream)
            print(file=self.stream)

        print('</table>', file=self.stream)
        return self

    def print_line(self, func):
        path, line_number, function_name = func

        base_dir = str(settings.BASE_DIR)
        should_bold = base_dir in path and '/site-packages/' not in path
        nice_path = path.replace(base_dir, '')
        nice_path = strip_extra_path(nice_path, '/site-packages')
        nice_path = strip_extra_path(nice_path, '/Python.framework/Versions')

        if should_bold:
            print('<tr class="own">', file=self.stream)
        else:
            print('<tr>', file=self.stream)

        def f8(x):
            return "%8.3f" % x

        cc, nc, tt, ct, callers = self.stats[func]
        c = str(nc)
        if nc != cc:
            c = c + '/' + str(cc)
        print(f'<td class="numeric">{c}</td>', file=self.stream)
        print(f'<td class="numeric">{f8(tt)}</td>', file=self.stream)
        if nc == 0:
            print('<td></td>', file=self.stream)
        else:
            print(f'<td>{f8(tt/nc)}</td>', file=self.stream)
        print(f'<td class="numeric">{f8(ct)}</td>', file=self.stream)
        if cc == 0:
            print('<td></td>', file=self.stream)
        else:
            print(f'<td class="numeric">{f8(ct/cc)}</td>', file=self.stream)

        if line_number and path:
            print(
                f'<td><a href="{src_debug_url_builder(path, line_number)}">{escape(function_name)}</a></td>',
                file=self.stream,
            )
        else:
            print(f'<td>{escape(function_name)}</td>', file=self.stream)

        from iommi import traversable
        if function_name in traversable.worst_offenders_candidates:
            print(f'<td><a href="?_iommi_func_worst_offender={escape(function_name)}">Worst offenders</a></td>', file=self.stream)
        else:
            print('<td></td>', file=self.stream)

        print(f'<td>{nice_path}</td>', file=self.stream)
        print(f'<td class="numeric">{line_number}</td>', file=self.stream)
        print('</tr>', file=self.stream)


class Middleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Disable profiling early on /media requests since touching request.user will add a
        # "Vary: Cookie" header to the response.
        request.profiler_disabled = False
        for prefix in MEDIA_PREFIXES:
            if request.path.startswith(prefix):
                request.profiler_disabled = True
                break

        if not should_profile(request):
            return self.get_response(request)

        prof = cProfile.Profile()
        prof.enable()
        request._iommi_prof = [prof]

        response = self.get_response(request)

        if not isinstance(response, HttpResponseBase):
            assert False, f'Got a response of type {type(response)}, expected an HttpResponse object. Middlewares are in the wrong order.'

        if getattr(request, '_iommi_func_worst_offender', None):
            s = StringIO()
            for stack, count in sorted(request._iommi_func_worst_offender.items(), key=lambda x: -x[1]):
                if count <= 1:
                    break

                print(f'----- {count} -----', file=s)
                print(stack, file=s)
                print(file=s)

            return HttpResponse(s.getvalue(), content_type='text/plain')

        if request._iommi_prof:
            if isinstance(response, StreamingHttpResponse):
                # consume the entire streaming response, redirecting to stdout
                for line in response.streaming_content:
                    print(line.decode(), file=sys.__stdout__)

            response = HttpResponse()
            for prof in request._iommi_prof:
                prof.disable()

            s = StringIO()
            ps = HTMLStats(*request._iommi_prof, stream=s)

            prof_command = request.GET.get('_iommi_prof')

            if prof_command == 'graph':
                with NamedTemporaryFile() as stats_dump:
                    ps.stream = stats_dump
                    ps.dump_stats(stats_dump.name)

                    gprof2dot_path = Path(sys.executable).parent / 'gprof2dot'
                    if not gprof2dot_path.exists():
                        raise Exception('gprof2dot not found. Please install it to use the graph feature.')

                    with subprocess.Popen(
                        (sys.executable, gprof2dot_path, '-f', 'pstats', stats_dump.name), stdout=subprocess.PIPE
                    ) as gprof2dot:
                        response['Content-Type'] = 'image/svg+xml'

                        dot_path = get_dot_path()
                        if dot_path:
                            response.content = subprocess.check_output((dot_path, '-Tsvg'), stdin=gprof2dot.stdout)
                        else:
                            response['Content-Type'] = 'text/plain'
                            response['Content-Disposition'] = "attachment; filename=gprof2dot-graph.txt"
                            response.content = subprocess.check_output('tee', stdin=gprof2dot.stdout)

                        gprof2dot.wait()

            elif prof_command == 'snake':
                # noinspection PyPackageRequirements
                try:
                    import snakeviz  # noqa
                except ImportError:
                    return HttpResponse('You must `pip install snakeviz` to use this feature')

                with NamedTemporaryFile() as stats_dump:
                    ps.stream = stats_dump
                    ps.dump_stats(stats_dump.name)

                    subprocess.Popen(
                        [sys.executable, str(Path(sys.executable).parent / 'snakeviz'), stats_dump.name],
                        stdin=None,
                        stdout=None,
                        stderr=None,
                    )

                    # We need to wait a bit to give snakeviz time to read the file
                    from time import sleep

                    sleep(3)

                return HttpResponse(
                    'You should have gotten a new browser window with snakeviz opened to the profile data'
                )

            else:
                ps = ps.sort_stats(prof_command or 'cumulative')
                ps.print_stats()

                result = s.getvalue()

                # language=html
                start_html = '''
                    <style>
                        html {
                            font-family: monospace;
                            white-space: pre-line;
                        }

                        div, table {
                            white-space: normal;
                        }

                        td, th {
                            white-space: nowrap;
                            padding-right: 0.5rem;
                            color: #666;
                        }

                        th {
                            text-align: left;
                        }

                        .numeric {
                            text-align: right;
                        }

                        .own td {
                            font-weight: bold;
                            color: black;
                        }

                        @media (prefers-color-scheme: dark) {
                            html {
                                background-color: black;
                                color: #bbb;
                            }
                            td, th {
                                color: #888;
                            }

                            .own td {
                                color: white;
                            }

                            a {
                                color: #1d5aff;
                            }
                            a:visited {
                                color: #681dff;
                            }
                        }
                    </style>

                    <div>
                        <a href="?_iommi_prof=graph">graph</a>
                        <a href="?_iommi_prof=snake">snakeviz</a>
                    </div>

                    <p></p>
                '''

                response.content = start_html.strip() + result

                response['Content-Type'] = 'text/html'

        return response



import inspect
import sys
from os.path import (
    dirname,
    isabs,
    join,
)

from django.http import HttpResponse
from django.template.loader import get_template

from iommi._web_compat import (
    HttpRequest,
    Template,
    format_html,
    mark_safe,
)
from iommi.base import items
from iommi.struct import Struct

from ._web_compat import settings


def iommi_debug_on():
    return getattr(settings, 'IOMMI_DEBUG', settings.DEBUG)


def dunder_path__format(row, **_):
    if row.dunder_path is None:
        return ''
    prefix = row.dunder_path.rpartition('__')[0]
    return format_html(
        '<span class="full-path">{prefix}{separator}</span>{name}',
        prefix=prefix,
        separator='__' if prefix else '',
        name=row.name,
    )


def endpoint__debug_tree(endpoint, **_):
    root = endpoint.iommi_root()
    assert root._is_bound

    def rows(node, name='', path=None):
        from iommi.member import (
            MemberBinder,
            Members,
        )
        from iommi.traversable import (
            PathNotFoundException,
            Traversable,
            declared_members,
        )

        if path is None:
            path = []
        is_struct = type(node) is Struct
        is_bound = getattr(node, '_is_bound', False)

        try:
            p = node.iommi_path if is_bound else None
        except PathNotFoundException:
            p = None

        type_name = type(node).__name__ if not is_struct else None
        prefix = ''
        if isinstance(node, Members):
            if node._declared_members:
                type_name = node._cls.__name__
                prefix = 'Contains '
            else:
                return

        children = []
        if isinstance(node, dict):
            children = list(node.items())
        elif isinstance(node, Traversable):
            children = []
            for k, v in items(declared_members(node)):
                if node._bound_members is not None and isinstance(node._bound_members.get(k, None), MemberBinder):
                    node._bound_members[k]._force_bind_all()
                children.append((k, (node._bound_members or {}).get(k, v)))

        if (isinstance(node, Members) or isinstance(node, dict)) and not children:
            return

        display_name = type_name
        shortcut_stack = getattr(node, 'iommi_shortcut_stack', None)
        shortcut = shortcut_stack[0] if shortcut_stack else None
        if shortcut:
            display_name = f'{type_name}.{shortcut}'
        yield Struct(
            name=name,
            obj=node,
            prefix=prefix,
            type_name=type_name,
            display_name=display_name,
            shortcut=shortcut,
            path=p,
            dunder_path='__'.join(path),
            included=is_bound,
        )

        for k, v in children:
            yield from rows(v, name=k, path=path + [k])

    from iommi import (
        Column,
        Table,
    )

    class TreeTable(Table):
        class Meta:
            template = Template(
                """
                <style>
                    .full-path {
                        opacity: 0.0;
                    }
                    tr:hover .full-path {
                        opacity: 0.6;
                    }

                    tr {
                        opacity: 0.4;
                    }
                    tr.included {
                        opacity: 1;
                    }

                </style>

                {% include "iommi/table/table.html" %}
            """
            )
            sortable = False
            page_size = None

            @staticmethod
            def row__attrs__class__included(row, **_):
                return row.included

        dunder_path = Column(
            cell__value=lambda row, **_: row.dunder_path,
            cell__format=dunder_path__format,
        )
        path = Column()
        type = Column(
            attr='type_name',
            cell__format=lambda row, **_: (
                format_html(
                    '{prefix}<a href="https://docs.iommi.rocks//{type_name}.html{anchor}">{display_name}</a>',
                    prefix=row.prefix,
                    type_name=row.type_name,
                    anchor=f'#{row.shortcut.replace("_", "-")}' if row.shortcut else '',
                    display_name=row.display_name,
                )
                if row.type_name
                else ''
            ),
        )
        included = Column.boolean()

    request = HttpRequest()
    request.method = 'GET'
    request.META = root._request.META
    if hasattr(root._request, 'user'):
        request.user = root._request.user

    return TreeTable(rows=rows(root)).bind(request=request)


def endpoint__debug_templates_used(endpoint, request, **_):
    root = endpoint.iommi_root()
    from iommi.part import render_root

    render_root(part=root)

    def url_for_template(t):
        if isinstance(t, str):
            t = get_template(t).origin.name
        else:
            t = t.origin.name

        return local_debug_url_builder(t, 1)

    unique_templates = list(dict.fromkeys(request.iommi_used_templates))
    links = [
        format_html(
            '<li><a href="{}">{}</a></li>',
            url_for_template(t),
            t,
        )
        for t in unique_templates
    ]
    links = format_html('{}' * len(links), *links)
    return HttpResponse(
        # language=HTML
        format_html(
            '''
                <html>
                    <head>
                        <style>
                        </style>
                    </head>
                    <body>
                        <ul>
                            {}
                        </ul>
                    </body>
                </html>
            ''',
            links,
        )
    )


def local_debug_url_builder(filename, lineno):
    if not isabs(filename):
        filename = join(settings.BASE_DIR, filename)
    if hasattr(settings, 'IOMMI_DEBUG_URL_MAPPING'):
        filename = filename.replace(*settings.IOMMI_DEBUG_URL_MAPPING)
    return f"pycharm://open?file={filename}" + ('' if lineno is None else f"&line={lineno:d}")


def src_debug_url_builder(filename, lineno=None):
    debug_url_builder = getattr(settings, 'IOMMI_DEBUG_URL_BUILDER', local_debug_url_builder)
    return debug_url_builder(filename, lineno)


def should_ignore_frame(frame, env_paths):
    assert isinstance(env_paths, set)
    module_name = frame.f_globals.get('__name__')

    if not module_name:
        return False

    if module_name.startswith('iommi.admin'):
        return True

    if module_name.startswith('_pydev_bundle.'):
        return True

    base_module_name = module_name.partition('.')[0]
    if base_module_name in (
        'iommi.declarative',
        'iommi',
        'django',
    ):
        if not module_name.endswith('__tests'):
            return True

    for env_path in env_paths:
        if frame.f_code.co_filename.startswith(env_path):
            return True

    if frame.f_code.co_filename == '<string>':
        return True

    if join('helpers', 'pycharm') in frame.f_code.co_filename:
        return True

    return False


def get_instantiated_at_info(frame):
    import os

    env_paths = {dirname(os.__file__), dirname(dirname(sys.executable))}

    for _ in range(100):
        frame = frame.f_back
        if frame is None:
            break

        if should_ignore_frame(frame, env_paths):
            continue

        return frame.f_code.co_filename, frame.f_lineno
    return None, None


def filename_and_line_num_from_part(part):
    from iommi.traversable import Traversable
    filename, line_num = getattr(part, '_instantiated_at_info', (None, None))
    if (filename, line_num) == (None, None) and isinstance(part, Traversable):
        # inspect.findsource() parses the full AST of the file, so beware the performance implications
        filename = inspect.getsourcefile(part.__class__)
        line_num = inspect.findsource(part.__class__)[1] + 1
        setattr(part, '_instantiated_at_info', (filename, line_num))
    return filename, line_num


def iommi_debug_panel(part):
    source_url = source_url_from_part(part)
    # language=js
    script = r"""
        window.iommi_start_pick = function() {
            window.iommi_pick_stack = [];

            function create(html) {
                let r = document.createElement('div');
                r.innerHTML = html;
                return r.firstChild;
            }

            window.iommi_close_pick_toolbar = function() {
                window.iommi_pick_stack.forEach(function(el) {
                    el[3].style.backgroundColor = el[2];
                });
                document.getElementById('iommi-pick-toolbar').remove()
            };

            function update_toolbar() {
                let toolbar = document.getElementById('iommi-pick-toolbar');
                if (!toolbar) {
                    return;
                }

                while(toolbar.firstChild) {
                    toolbar.removeChild(toolbar.firstChild);
                }

                toolbar.append(create('<div style="float: right" onclick="iommi_close_pick_toolbar()">close</div>'));
                for (let i in window.iommi_pick_stack) {
                    let x = window.iommi_pick_stack[i];
                    toolbar.append(create('<div style="background-color: ' + getColor(i) + '">' + x[0] + ' <a href="https://docs.iommi.rocks//' + x[1] + '.html">' + x[1] + '</a></div>'));
                }
            }

            let with_iommi_path = document.querySelectorAll('*[data-iommi-path]');

            let colors = [
                'rgb(255, 255, 191)',
                'rgb(254, 224, 139)',
                'rgb(253, 174,  97)',
                'rgb(244, 109,  67)',
                'rgb(213,  62,  79)',
                'rgb(158,   1,  66)',
                'rgb(230, 245, 152)',
                'rgb(171, 221, 164)',
                'rgb(102, 194, 165)',
                'rgb( 50, 136, 189)',
                'rgb( 94,  79, 162)',
            ];

            function getColor(index) {
                return colors[Math.min(index, colors.length - 1)]
            }

            function mouseenter() {
                window.iommi_pick_stack.push([this.getAttribute('data-iommi-path'), this.getAttribute('data-iommi-type'), this.style.backgroundColor, this])
                this.style.backgroundColor = getColor(window.iommi_pick_stack.length-1);
                update_toolbar();
            }

            function mouseleave() {
                if (window.iommi_pick_stack.length) {
                    this.style.backgroundColor = window.iommi_pick_stack.pop()[2];
                    update_toolbar();
                }
            }
            function click() {
                document.querySelectorAll('*[data-iommi-path]').forEach(function (e) {
                    e.removeEventListener('mouseenter', mouseenter)
                    e.removeEventListener('mouseleave', mouseleave)
                    e.removeEventListener('click', click)
                });
            }

            with_iommi_path.forEach(function (e) {
                e.addEventListener('mouseenter', mouseenter);
                e.addEventListener('mouseleave', mouseleave);
                setTimeout(function(){
                    e.addEventListener('click', click);
                });
            });

            let toolbar = create('<div id="iommi-pick-toolbar" style="position: fixed; left: 0; bottom: 0; width: 100%; background-color: white; color: black; padding: 4px; border-top: 2px solid #1084ff; z-index: 200">');

            document.getElementsByTagName('body')[0].append(toolbar);
        };
    """

    from iommi.menu import get_debug_menu

    return get_debug_menu(sub_menu__code__url=source_url).bind(request=part.get_request()).__html__() + mark_safe(
        f'<script>{script}</script>'
    )


def source_url_from_part(part):
    filename, lineno = filename_and_line_num_from_part(part)
    if filename is not None:
        source_url = src_debug_url_builder(filename, lineno)
    else:
        source_url = None
    return source_url



from django.db.models import PositiveIntegerField
from django import VERSION as DJANGO_VERSION


class SortOrderField(PositiveIntegerField):
    def __init__(self, to=None, **kwargs):
        kwargs.setdefault("null", False)
        kwargs.setdefault("blank", False)
        if DJANGO_VERSION[0] >= 5:
            kwargs.setdefault("db_default", 0)
        kwargs.setdefault("db_index", True)
        kwargs.setdefault("default", 0)
        kwargs.setdefault("editable", False)
        super().__init__(to, **kwargs)



import warnings
from itertools import groupby
from typing import (
    Callable,
    Dict,
    List,
    Tuple,
)

from django.utils.translation import gettext_lazy

from iommi._web_compat import (
    format_html,
    slugify,
)
from iommi.attrs import (
    Attrs,
    render_attrs,
)
from iommi.base import (
    capitalize,
    values,
)
from iommi.declarative.namespace import EMPTY
from iommi.fragment import (
    Fragment,
    Tag,
)
from iommi.member import Members
from iommi.part import Part
from iommi.refinable import (
    EvaluatedRefinable,
    Prio,
    Refinable,
    SpecialEvaluatedRefinable,
)
from iommi.shortcut import with_defaults


class Action(Fragment):
    # language=rst
    """
    The `Action` class describes buttons and links.

    Examples:

    .. code-block:: python

        # @test
        actions = dict(
        # @end

        # Link
        example=Action(attrs__href='http://example.com'),

        # Link with icon
        edit=Action.icon('pencil-square', attrs__href="edit/"),

        # Button
        button=Action.button(display_name='Button title!'),

        # A submit button
        submit=Action.submit(display_name='Do this'),

        # The primary submit button on a form.
        primary=Action.primary(),

        # @test
        )

        from iommi.refinable import Prio
        foo = Page(
            parts__css=html.style('''
                a, button {
                    display: block !important;
                    margin-bottom: 0.5rem;
                }
            '''),
            parts=actions,
        )
        show_output(foo)
        # @end

    Notice that because forms
    with a single primary submit button are so common, iommi assumes
    that if you have an action called submit and do NOT explicitly
    specify the action that it is a primary action. This is only
    done for the action called `submit`, inside the Forms actions
    `Namespace`.

    For that reason this works:

    .. code-block:: python

        class MyForm(Form):
            class Meta:
                @staticmethod
                def actions__submit__post_handler(form, **_):
                    if not form.is_valid():
                        return

                    ...

    and is roughly equivalent to

    .. code-block:: python

        def on_submit(form, **_):
            if not form.is_valid():
                return

        class MyOtherForm(Form):
            class Meta:
                actions__submit = Action.primary(post_handler=on_submit)

        # @test
        r = req('post', **{'-submit': ''})
        MyForm().bind(request=r).render_to_response()
        MyOtherForm().bind(request=r).render_to_response()
        # @end
    """

    group: str = EvaluatedRefinable()
    display_name: str = EvaluatedRefinable()
    post_handler: Callable = Refinable()

    @with_defaults(
        tag='a',
        display_name=lambda action, **_: capitalize(action._name).replace('_', ' '),
    )
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_refine_done(self):
        if self.tag == 'input':
            if self.display_name and 'value' not in self.attrs:
                self.attrs.value = self.display_name
        else:
            self.children['text'] = self.display_name
            if self.tag == 'button' and 'value' in self.attrs:
                assert False, 'You passed attrs__value, but you should pass display_name'
        super().on_refine_done()

    def on_bind(self):
        super().on_bind()

    def own_evaluate_parameters(self):
        return dict(action=self)

    def __repr__(self):
        return Part.__repr__(self)

    def own_target_marker(self):
        return f'-{self.iommi_path}'

    def is_target(self):
        return self.own_target_marker() in self.iommi_parent().iommi_parent()._request_data

    @classmethod
    @with_defaults(
        tag='button',
    )
    def button(cls, **kwargs):
        return cls(**kwargs)

    @classmethod
    @with_defaults(
        attrs__accesskey='s',
        attrs__name=lambda action, **_: action.own_target_marker(),
        display_name=gettext_lazy('Submit'),
    )
    def submit(cls, **kwargs):
        return cls.button(**kwargs)

    @classmethod
    @with_defaults
    def primary(cls, **kwargs):
        return cls.submit(**kwargs)

    @classmethod
    @with_defaults
    def delete(cls, **kwargs):
        return cls.submit(**kwargs)

    @classmethod
    @with_defaults(
        extra__icon_attrs__class=EMPTY,
        extra__icon_attrs__style=EMPTY,
    )
    def icon(cls, icon, *, display_name=None, icon_classes=None, **kwargs):
        if icon_classes is not None:
            assert False, 'icon_classes is removed, use the extra__icon_attrs__class namespace'
        return cls(**kwargs).refine(
            extra__icon=icon,
            extra__orig_display_name=display_name,
            display_name=default_action__icon__display_name,
            prio=Prio.shortcut,
        )


def default_action__icon__display_name(action, **_):
    if not action.extra.get('icon', None):
        return action.extra.orig_display_name

    attrs = action.extra.icon_attrs
    attrs['class'][action.extra.get('icon_prefix', '') + action.extra.icon] = True

    return format_html('<i{}></i> {}', render_attrs(attrs), action.extra.orig_display_name or capitalize(action._name).replace('_', ' '))


def group_actions(actions: Dict[str, Action]):
    actions_with_group = (action for action in values(actions) if action.group is not None)

    grouped_actions: List[Tuple[str, str, List[Action]]] = [
        (group_name, slugify(group_name), list(actions_in_group))
        for group_name, actions_in_group in groupby(actions_with_group, key=lambda x: x.group)
    ]

    for _, _, actions_in_group in grouped_actions:
        for action in actions_in_group:
            action.attrs.role = 'menuitem'
            action.attrs['class']['dropdown-item'] = True

    actions_without_group = [action for action in values(actions) if action.group is None]
    return actions_without_group, grouped_actions


class Actions(Members, Tag):
    attrs: Attrs = SpecialEvaluatedRefinable()
    tag = EvaluatedRefinable()
